#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import configparser
from os import path

from auth0_login.logging import fatal


class __Setting(object):
    def __init__(self):
        self.filename = '.pcke-login'
        self.__SECTION = 'DEFAULT'

    @property
    def filename(self):
        return self.__filename

    @filename.setter
    def filename(self, filename):
        self.__filename = filename
        self.config = configparser.ConfigParser()
        self.config.read([path.expanduser(path.expandvars(f'~/{filename}')), f'{filename}'])

    @property
    def SECTION(self):
        return self.__SECTION

    @SECTION.setter
    def SECTION(self, section):
        self.__SECTION = section

    @property
    def attributes(self) -> dict:
        return {v[0]: v[1] for v in self.config.items(self.SECTION)}

    @property
    def LISTEN_PORT(self):
        return self.config.getint(self.SECTION, 'listen_port', fallback=12200)

    @property
    def ROLE_DURATION(self):
        return self.config.getint(self.SECTION, 'role_duration', fallback=3600)

    @property
    def CLIENT_ID(self):
        if not self.config.has_option(self.SECTION, 'client_id'):
            fatal(f'property client_id is missing from ~/{self.filename}')
        return self.config.get(self.SECTION, 'client_id')

    @property
    def IDP_URL(self):
        if not self.config.has_option(self.SECTION, 'idp_url'):
            fatal(f'property client_id is missing from ~/{self.filename}')
        return self.config.get(self.SECTION, 'idp_url')

    @property
    def exists(self):
        return self.SECTION == 'DEFAULT' or self.config.has_section(self.SECTION)


setting = __Setting()



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import hashlib
import json
import logging
import webbrowser
from base64 import urlsafe_b64encode
from http.server import HTTPServer
from random import getrandbits
from sys import stdout
from urllib.parse import urlencode
from uuid import uuid4

import click

from auth0_login import fatal, setting
from auth0_login.pkce.callback import PKCEAccessTokenCallbackHandler
from auth0_login.util import assert_listen_port_is_available, get_listen_port_from_url


class PKCEGetIdTokenCommand(object):
    """
    requests an JWT id token using the PKCE authorization flow and prints
    all the returned data to standard output.

    The request is sent  `{idp_url}/authorize`, the callback
    defaults to `http://localhost:{listen_port}/callback`, but may be
    explicitly set using the `pcke_callback_url` property.

    """

    def __init__(self):
        self.client_id = setting.CLIENT_ID
        self.scope = "openid profile"
        self.tokens = {}
        self.state = str(uuid4())
        self.verifier = self.b64encode(bytearray(getrandbits(8) for _ in range(32)))
        self.challenge = self.b64encode(hashlib.sha256(self.verifier.encode('ascii')).digest())
        self.callback_url = setting.attributes.get('pkce_callback_url',
                                                   f'http://localhost:{setting.LISTEN_PORT}/callback')

    @property
    def token_url(self):
        return f'{setting.IDP_URL}/oauth/token'

    @property
    def authorize_url(self):
        return f'{setting.IDP_URL}/authorize'

    @property
    def listen_port(self):
        return get_listen_port_from_url(self.callback_url)

    def set_tokens(self, tokens):
        self.tokens = tokens

    def accept_access_code(self):
        PKCEAccessTokenCallbackHandler.client_id = setting.CLIENT_ID
        PKCEAccessTokenCallbackHandler.token_url = self.token_url
        PKCEAccessTokenCallbackHandler.callback_url = self.callback_url
        PKCEAccessTokenCallbackHandler.verifier = self.verifier
        PKCEAccessTokenCallbackHandler.state = self.state
        PKCEAccessTokenCallbackHandler.handler = (lambda tokens: self.set_tokens(tokens))
        httpd = HTTPServer(('0.0.0.0', self.listen_port), PKCEAccessTokenCallbackHandler)
        try:
            httpd.handle_request()
        finally:
            httpd.server_close()

    @property
    def query_parameters(self):
        return {
            "response_type": "code",
            "scope": self.scope,
            "client_id": self.client_id,
            "code_challenge": self.challenge,
            "code_challenge_method": "S256",
            "redirect_uri": self.callback_url,
            "state": self.state
        }

    @property
    def url(self):
        return self.authorize_url + '?' + urlencode(self.query_parameters)

    def request_authorization(self):
        logging.debug('url = %s', self.url)
        webbrowser.open(self.url)
        self.accept_access_code()

    @staticmethod
    def b64encode(s):
        return urlsafe_b64encode(s).decode('ascii').strip("=")

    def run(self):
        assert_listen_port_is_available(self.listen_port)
        self.request_authorization()
        if self.tokens:
            json.dump(self.tokens, stdout)
        else:
            fatal('no token retrieved')


class PKCEGetAccessTokenCommand(PKCEGetIdTokenCommand):
    """
    requests an JWT access token using the PKCE authorization flow for the
    specified `audience` and `scope`. All returned data to printed to
    standard output.

    Both `audience` and `scope` can be specified as a command line option
    or in the .oauth-cli.ini.

    The request is sent  `{idp_url}/authorize`, the callback
    defaults to `http://localhost:{listen_port}/callback, but may be
    explicitly set using the `pcke_callback_url` property.

    """

    def __init__(self):
        super(PKCEGetAccessTokenCommand, self).__init__()
        self.audience = setting.attributes.get('audience')
        self.scope = setting.attributes.get('scope', 'openid profile')

    @property
    def query_parameters(self):
        result = super(PKCEGetAccessTokenCommand, self).query_parameters
        result.update({"audience": self.audience, "scope": self.scope})
        return result

    def run(self):
        if not self.audience:
            fatal('audience is required')
        super(PKCEGetAccessTokenCommand, self).run()


@click.command('get-access-jwt', help=PKCEGetAccessTokenCommand.__doc__)
@click.option('--audience', help='to obtain an access token for. default from ~/.oauth-cli.ini')
@click.option('--scope', help='of the access token')
def get_access_token(audience, scope):
    cmd = PKCEGetAccessTokenCommand()
    if audience:
        cmd.audience = audience
    if scope:
        cmd.scope = scope
    cmd.run()


@click.command('get-id-jwt', help=PKCEGetIdTokenCommand.__doc__)
def get_id_token():
    cmd = PKCEGetIdTokenCommand()
    cmd.run()



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from sys import stdout

import click

from auth0_login import fatal, setting
from auth0_login.aws.console import open_aws_console
from auth0_login.aws.credentials import write_aws_credentials
from auth0_login.aws.account import aws_accounts
from auth0_login.aws.saml_assertion import AWSSAMLAssertion
from auth0_login.saml import SAMLGetAccessTokenCommand


class AWSSTSGetCredentialsFromSAMLCommand(SAMLGetAccessTokenCommand):
    """
    get AWS credentials using the obtained SAML token and
    stores then in `~/.aws/credentials`.

    As multiple AWS roles may have been granted to the SAML token,
    the caller has to specify the `account` number and `role` name to
    generate the credentials for. If you are unsure which accounts
    and roles have been granted, use the `--show` option

    The credentials will be stored under the specified `profile` name.
    By specifying `--open-console` it will open the AWS console too.
    """

    def __init__(self, account, role, profile):
        super(AWSSTSGetCredentialsFromSAMLCommand, self).__init__()
        self.account = account if account else setting.attributes.get('aws_account')
        if not account and self.account:
            self.account = aws_accounts.get_account(self.account).number

        self.role = role if role else setting.attributes.get('aws_role')
        self.profile = profile if profile else setting.attributes.get('aws_profile')
        self.open_console = setting.attributes.get('aws_console', False)
        self.saml_response: AWSSAMLAssertion = None

    def set_saml_response(self, saml_response):
        self.saml_response = AWSSAMLAssertion(saml_response)

    def print_roles(self):
        for role in self.saml_response.available_roles():
            account = aws_accounts.get_account(role.account)
            stdout.write(f'[{role.name}@{account.alias}]\n')
            stdout.write(f'idp_url = {setting.IDP_URL}\n')
            stdout.write(f'client_id = {setting.CLIENT_ID}\n')
            stdout.write(f'aws_account = {account.alias}\n')
            stdout.write(f'aws_role = {role.name}\n')
            stdout.write(f'aws_profile = {role.name}@{account.alias}\n\n')

    def show_account_roles(self):
        self.request_authorization()
        self.print_roles()

    @property
    def role_arn(self):
        return f'arn:aws:iam::{self.account}:role/{self.role}'

    def run(self):
        if not (self.account and self.role and self.profile):
            fatal('--account, --role and --profile are required.')

        self.request_authorization()

        credentials = self.saml_response.assume_role(self.role_arn, setting.ROLE_DURATION)
        write_aws_credentials(credentials, self.profile)
        if self.open_console:
            open_aws_console(self.profile)


@click.command('aws-assume-role', help=AWSSTSGetCredentialsFromSAMLCommand.__doc__)
@click.option('--account', help='aws account number or alias')
@click.option('--role', help='to assume using the token')
@click.option('--profile', help='to store the credentials under')
@click.option('--show', is_flag=True, default=False, help='account roles available to assume')
@click.option('--open-console', '-C', count=True, help=' after credential refresh')
def assume_role_with_saml(account, role, profile, show, open_console):

    aws_account = aws_accounts.get_account(account).number if account else None
    cmd = AWSSTSGetCredentialsFromSAMLCommand(aws_account, role, profile)
    if show:
        cmd.show_account_roles()
    else:
        if open_console:
            cmd.open_console = True
        cmd.run()



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import click
import logging
import webbrowser

import requests
from boto3 import Session
from botocore.credentials import ReadOnlyCredentials
from botocore.exceptions import ClientError

from auth0_login import fatal, setting


def get_federated_credentials(session: Session) -> ReadOnlyCredentials:
    iam = session.client('iam')
    sts = session.client('sts')
    policy = {"Version": "2012-10-17", "Statement": [{"Action": "*", "Effect": "Allow", "Resource": "*"}]}
    try:
        user = iam.get_user()
        r = sts.get_federation_token(Name=user['User']['UserName'], DurationSeconds=setting.ROLE_DURATION, Policy=json.dumps(policy))
        c = r['Credentials']
        return ReadOnlyCredentials(access_key=c['AccessKeyId'], secret_key=c['SecretAccessKey'], token=c['SessionToken'])
    except ClientError as e:
        fatal('failed to get federation token, %s', e)

def open_aws_console(profile: str):
    """
    opens the AWS console for the specified profile.
    """
    s: Session = Session(profile_name=profile)
    c: ReadOnlyCredentials = s.get_credentials().get_frozen_credentials()
    if not c.token:
        logging.debug('getting federated credentials')
        c = get_federated_credentials(s)

    if not c.token:
        fatal('cannot generated a console signin URL from credentials without a session token')

    creds = {'sessionId': c.access_key, 'sessionKey': c.secret_key, 'sessionToken': c.token}
    logging.debug('obtaining AWS console signin token')
    response = requests.get("https://signin.aws.amazon.com/federation",
                            params={'Action': 'getSigninToken',
                                    'SessionType': 'json', 'Session': json.dumps(creds)})
    if response.status_code != 200:
        fatal("could not generate Console signin URL, %s,\n%s", response.status_code, response.text)

    signin_token = response.json()['SigninToken']
    params = {'Action': 'login', 'Issuer': 'awslogin', 'Destination': 'https://console.aws.amazon.com/',
              'SigninToken': signin_token}
    logging.debug('opening AWS console')
    console = requests.Request('GET', 'https://signin.aws.amazon.com/federation', params=params)
    prepared_link = console.prepare()
    webbrowser.open(prepared_link.url)


@click.command('aws-console', help='open AWS console from profile')
@click.option('--verbose', is_flag=True, default=False, help=' for tracing purposes')
@click.option('--profile', required=True, help='to store the credentials under')
def main(verbose, profile):
    logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if verbose else logging.INFO))
    open_aws_console(profile)
